/* IPLimit (C)oderite 1999 by Frank DENIS aka Jedi/Sector One <j@4u.net> */

/* Customize these settings */

#define VARFILE "/var/run/iplimit.dat" /* Must be here, even if empty */
#define DENY_MSG _("Stop that, lame flooder")

/* If your system needs <memory.h> or <sys/time.h>, add them here */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>

/* Don't edit below that line, or send your patches to <j@4u.net> */

#define _(X) X
#define MAX_IP_LEN 24
#define MAX_PORT_LEN 8

#ifndef O_NONBLOCK
# define O_NONBLOCK O_NDELAY
#endif

typedef struct Cnx_ {
   char ip[MAX_IP_LEN];
   char port[MAX_PORT_LEN];
   time_t stamp;
} Cnx;

typedef struct CnxListEntry_ {
   Cnx cnx;
   struct CnxListEntry_ *next;
} CnxListEntry;

static void slog(const char * const s)
{
   static char a;
   
   if (s == NULL) {
      if (a != 0) {
	 closelog();
      }
   } else {
      if (a == 0) {
	 a++;
#ifdef LOG_AUTHPRIV
	 openlog("IPLimit", LOG_NOWAIT, LOG_AUTHPRIV);
#else
	 openlog("IPLimit", LOG_NOWAIT, LOG_DAEMON);	 
#endif
      }
      syslog(LOG_WARNING, s);
   }
}

#define LOGLINE 250

static void iplog(const char * const remote_ip,
		  const char * const local_port,
		  const char * const s)
{
   static char line[LOGLINE];
   
   strncpy(line, "Service [", LOGLINE);
   strncat(line, local_port, LOGLINE - strlen(line));
   strncat(line, "] host [", LOGLINE - strlen(line));
   strncat(line, remote_ip, LOGLINE - strlen(line));
   strncat(line, "] ", LOGLINE - strlen(line));
   strncat(line, s, LOGLINE - strlen(line));
   slog(line);
}

int main(int argc, char *argv[])
{
   const char * const remote_ip = getenv("TCPREMOTEIP");
   const char * const local_port = getenv("TCPLOCALPORT");      
   int fp;
   Cnx tmp_cnx;
   CnxListEntry *cnx_list_first = NULL, *cnx_list_last = NULL;   
   const CnxListEntry *cnx_list_scan;
   int nb_cnx = 0, nb_cnx_total = 0;
   int max_per_host, max_per_service;
   double timeout;
   const time_t stamp = time(NULL);   
   
   if (argc < 5) {
      fputs(_("IPLimit 0.1 by Frank DENIS aka Jedi/Sector One <j@4u.net>\n")
	    _("Usage : iplimit <max per host> <max per service> <time> <program> [options]\n")
	    _("Ex : iplimit 30 15 30 /usr/sbin/in.identd -l -e -o\n\n"), stderr);
      return -5;
   }
   if ((max_per_host = atoi(argv[1])) <= 0 ||
       (max_per_service = atoi(argv[2])) <= 0 ||
       (timeout = strtod(argv[3], NULL)) <= 0.0) {
      fputs(_("Invalid limits\n"), stderr);
      return -6;
   }
   if (max_per_service > max_per_host) {
      fputs(_("Warning : max/service > max/host : that's stupid"), stderr);
   }
   if (remote_ip == NULL) {
      fputs(_("Remote IP not set ?!\n"), stderr);
      return -1;
   } else if (local_port == NULL) {
      fputs(_("Local port not set ?!\n"), stderr);
      return -2;
   }
   if ((fp = open(VARFILE, O_RDONLY | O_NONBLOCK)) < 0) {
      perror(_("Can't read " VARFILE));
      return -3;
   }
   while (read(fp, &tmp_cnx, sizeof tmp_cnx) > 0) {
      CnxListEntry *new_entry;

      if (difftime(stamp, tmp_cnx.stamp) > timeout) {
	 continue;
      }
      if ((new_entry = malloc(sizeof *new_entry)) == NULL) {
	 perror(_("Malloc"));
	 close(fp);
	 return -4;
      }
      new_entry->cnx = tmp_cnx;
      new_entry->next = NULL;
      if (cnx_list_first == NULL) {
	 cnx_list_first = cnx_list_last = new_entry;
      } else {
	 cnx_list_last->next = new_entry;
	 cnx_list_last = new_entry;
      }
   }
   cnx_list_scan = cnx_list_first;
   while (cnx_list_scan != NULL) {
      if (strcmp(cnx_list_scan->cnx.ip, remote_ip) == 0) {
	 nb_cnx_total++;
	 if (strcmp(cnx_list_scan->cnx.port, local_port) == 0) {
	    nb_cnx++;
	 }
      }
      cnx_list_scan = cnx_list_scan->next;
   }
   close(fp);
   if ((fp = open(VARFILE, O_WRONLY | O_NONBLOCK)) < 0) {
      perror(_("Can't create " VARFILE));
      return -3;
   }   
   if (flock(fp, LOCK_EX) < 0) {
      perror(_("Can't lock " VARFILE));
      return -3;
   }
   cnx_list_scan = cnx_list_first;
   while (cnx_list_scan != NULL) {
      write(fp, &cnx_list_scan->cnx, sizeof cnx_list_scan->cnx);
      cnx_list_scan = cnx_list_scan->next;
   }
   strncpy(tmp_cnx.ip, remote_ip, sizeof tmp_cnx.ip);
   strncpy(tmp_cnx.port, local_port, sizeof tmp_cnx.port);
   tmp_cnx.stamp = stamp;
   write(fp, &tmp_cnx, sizeof tmp_cnx);
   if (flock(fp, LOCK_UN) < 0) {
      perror(_("Can't unlock " VARFILE));
      return -3;
   }   
   close(fp);
   if (nb_cnx == max_per_service) {
      iplog(remote_ip, local_port, 
	    _("was rejected : too many connections to the same service"));
      goto xa;
   }
   if (nb_cnx_total == max_per_host) {
      iplog(remote_ip, local_port,
	    _("was rejected : too many connections to that host"));
      goto xa;
   }
   if (nb_cnx > max_per_service || nb_cnx_total > max_per_host) {
      xa:
      puts(DENY_MSG);
      return 1;
   }
   slog(NULL);
   execvp(argv[4], &argv[4]);
      
   return 0;
}
